// Copyright (c) Microsoft Corporation.  All rights reserved.

//------------------------------------------------------------------------------
// PLEASE NOTE:  Microsoft Corporation (or based on where you live, one of its affiliates) 
// licenses this supplement to you.  You may use it with each validly licensed copy of Microsoft 
// Visual Studio Pre-Release Code-Name "ORCAS" software (the “software”).  You may not use the 
// supplement if you do not have a license for the software.  The license terms for the software 
// apply to your use of this supplement. 
//
// Please note: As this software is distributed in Quebec, Canada, this notice is provided below in French.
// Remarque: Ce logiciel étant distribué au Québec, Canada, les termes de cette licence sont fournis ci-dessous en français.
// REMARQUE:  Microsoft Corporation (ou en fonction du lieu où vous vivez, l’un de ses affiliés) vous 
// accorde une licence pour ce supplément.  Vous pouvez utiliser une copie de ce supplément avec chaque 
// copie concédée sous licence du logiciel Microsoft Visual Studio Pre-Release Code-Name "ORCAS" (le « logiciel »).  
// Vous n’êtes pas autorisé à utiliser ce supplément si vous n’êtes pas titulaire d’une licence pour le logiciel.  
// Les conditions de licence pour le logiciel s’appliquent à l’utilisation de ce supplément.  
//------------------------------------------------------------------------------
using System;
using System.Data;
using System.Data.Common;
using System.Data.EntityClient;

class Program
{
    // Note the use of named connections in the connection string. This points
    // to the actual connection string in the config file
    const string CONNECTION_STRING = "Name=Northwind";
    
    static void Main(string[] args) {
        SimpleQuery();
        QueryWithParams();
        FetchWholeEntities();
        Inheritance_PolymorphicResultSets();
        Inheritance_FilterByType();
        RelationshipTraversal();
        RegularInnerJoin();
        CrossApply();
        CrossApplyWithAggregate();
        NestedResultSets();
        Aggregate();
    }

    // Execute a really simple query against the EDM schema to
    // fetch all the Category entities from the Categories entity set
    static void SimpleQuery() {
        using(EntityConnection con = new EntityConnection(CONNECTION_STRING)) {
            con.Open();

            EntityCommand cmd = new EntityCommand(
                "SELECT c.CategoryName, c.Description " +
                "FROM Northwind.Categories AS c " +
                "ORDER BY c.CategoryName",
                con);

            // Entity command requires SequentialAccess
            DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while(r.Read()) {
                Console.WriteLine("{0}\t{1}", r[0], r[1]);
            }
        }

        Separator();
    }

    // Use parameters in a query to include external values such as user input into a query
    static void QueryWithParams() {
        using(EntityConnection con = new EntityConnection(CONNECTION_STRING)) {
            con.Open();

            EntityCommand cmd = new EntityCommand(
                "SELECT c.CategoryName, c.Description " +
                "FROM Northwind.Categories AS c " +
                "WHERE c.CategoryID = @id",
                con);
            
            // parameters in the EntityCommand are like parameters in any other ADO.NET provider, just
            // create them through the parameters collection. The only difference is that parameter
            // names shouldn't include the "@" sign
            cmd.Parameters.AddWithValue("id", 1); // 1 is a hardcoded ID that's always there in the sample DB

            // Entity command requires SequentialAccess
            DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while(r.Read()) {
                Console.WriteLine("{0}\t{1}", r[0], r[1]);
            }
        }

        Separator();
    }

    // Use Entity SQL constructs to select whole entities instead of projecting
    // individual properties (columns)
    static void FetchWholeEntities() {
        using(EntityConnection con = new EntityConnection(CONNECTION_STRING)) {
            con.Open();

            // instead of projecting out individual members, we can request for the whole
            // entity to be fetched
            EntityCommand cmd = new EntityCommand(
                "SELECT c "+
                "FROM Northwind.Categories AS c",
                con);
            
            // Entity command requires SequentialAccess
            DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while(r.Read()) {
                // since we projected a single column "c" in the SELECT statement, we get a
                // single column back; this column has a structured value, which is the 
                // Category entity. So the type of the row will be DbDataRecord and the
                // record will contain a column for each member of the Category entity type.
                DbDataRecord category = (DbDataRecord)r["c"];

                Console.WriteLine("{0}\t{1}", category["CategoryName"], category["Description"]);
            }
            r.Close();

            Console.WriteLine();
            Console.WriteLine("Now without the record wrapper around each entity:");

            // when projecting a single column you can choose to eliminate the record
            // wrapper by using the "VALUE" Entity SQL keyword, for example:
            cmd.CommandText = "SELECT VALUE c FROM Northwind.Categories AS c";

            r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while(r.Read()) {
                // this time the record wrapper is not there, so each of the properties
                // of the Category entity is presented as a top-level column
                Console.WriteLine("{0}\t{1}", r["CategoryName"], r["Description"]);
            }
        }

        Separator();
    }

    // The Products entity-set has instances of the Product and DiscontinuedProduct
    // entity types (DiscontinuedProduct is a subtype of Product). Formulate a query
    // that returns instances of both and discriminate them in the client
    static void Inheritance_PolymorphicResultSets() {
        using(EntityConnection con = new EntityConnection(CONNECTION_STRING)) {
            con.Open();

            EntityCommand cmd = new EntityCommand(
                "SELECT VALUE p " +
                "FROM Northwind.Products AS p",
                con);
            
            // Entity command requires SequentialAccess
            DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while(r.Read()) {
                // if we're only accessing properties that are defined in the most base
                // type (e.g. Product in this case), then no special casing is needed; however,
                // here we want to access a property that's only available for a specific
                // subtype (the UnitsInStock property that's only in DiscontinuedProduct). So
                // we use the record metadata to find out the type of each entity

                // IExtendedDataRecord gets us access to the metadata information
                IExtendedDataRecord productRecord = (IExtendedDataRecord)r;

                // Print out properties that are present in all instances of Product or its subtypes
                Console.Write("{0}\t{1}", productRecord["ProductName"], productRecord["UnitPrice"]);

                // if it's DiscontinuedProduct then print out extra information, otherwise
                // just output a new line
                if(productRecord.DataRecordInfo.Metadata.EdmType.Name == "DiscontinuedProduct")
                    Console.WriteLine("\t{0}", r["UnitsInStock"]);
                else
                    Console.WriteLine();
            }
        }

        Separator();
    }

    // The Products entity-set has instances of the Product and DiscontinuedProduct
    // entity types (DiscontinuedProduct is a subtype of Product). Formulate a query
    // that returns only instances of DiscontinuedProduct or one of its subtypes
    static void Inheritance_FilterByType() {
        using(EntityConnection con = new EntityConnection(CONNECTION_STRING)) {
            con.Open();

            // Use the "IS OF" Entity SQL operator to filter entities in
            // the Products entity-set that are of type DiscontinuedProducts or
            // anyo of its subtypes
            EntityCommand cmd = new EntityCommand(
                "SELECT VALUE p " +
                "FROM Northwind.Products AS p " +
                "WHERE p IS OF (NorthwindLib.DiscontinuedProduct)",
                con);
            
            // Entoty command requires SequentialAccess
            DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while(r.Read()) {
                // Since the query will return only DiscontinuedProduct instances, we
                // can access the UnitsInStock property without checking for the type
                // in the client
                Console.WriteLine("{0}\t{1}\t{2}", r["ProductName"], r["UnitPrice"], r["UnitsInStock"]);
            }
        }

        Separator();
    }

    // Use the NAVIGATE operator to traverse the relationship between categories and
    // products that's described in the model (the association is called "Category_Product"
    // in the EDM model)
    static void RelationshipTraversal() {
        using(EntityConnection con = new EntityConnection(CONNECTION_STRING)) {
            con.Open();

            // Find all products from the "Beverages" category 
            // NOTE: we're actually looking at supporting using "." for traversing
            // relationships instead of having to use the NAVIGATE operator
            EntityCommand cmd = new EntityCommand(
                "SELECT VALUE p " +
                "FROM Northwind.Products AS p " +
                "WHERE NAVIGATE(p, NorthwindLib.Category_Product).CategoryName = @name",
                con);
            cmd.Parameters.AddWithValue("name", "Beverages");

            // Entity command requires SequentialAccess
            DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while(r.Read()) {
                Console.WriteLine("{0}", r["ProductName"]);
            }
        }

        Separator();
    }

    // Use the CROSS APPLY operator in the FROM clause as the equivalent of INNER JOIN when
    // an association from the entity on the left is defined to the entity on the right.
    static void CrossApply()
    {
        using (EntityConnection con = new EntityConnection(CONNECTION_STRING))
        {
            con.Open();

            // Find all products in each category category
            EntityCommand cmd = new EntityCommand(
                "SELECT c.CategoryName, p.ProductName " +
                "FROM Northwind.Categories AS c " +
                "CROSS APPLY c.Products AS p ",
                con );

            // Entity command requires SequentialAccess
            DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while (r.Read())
            {
                Console.WriteLine("{0}\t{1}", r["CategoryName"], r["ProductName"]);
            }
        }

        Separator();
    }

    // Use the CROSS APPLY operator in the FROM clause as the equivalent of INNER JOIN when
    // an association from the entity on the left is defined to the entity on the right, with
    // additional aggregation.
    static void CrossApplyWithAggregate()
    {
        using (EntityConnection con = new EntityConnection(CONNECTION_STRING))
        {
            con.Open();

            // Find the average unit price of products in each category category
            EntityCommand cmd = new EntityCommand(
                "SELECT c.CategoryName, SqlServer.AVG(p.UnitPrice) as aup " +
                "FROM Northwind.Categories AS c " +
                "CROSS APPLY c.Products AS p " +
                "GROUP BY c.CategoryName",
                con);

            // Entity command requires SequentialAccess
            DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while (r.Read())
            {
                Console.WriteLine("{0}\t{1}", r["CategoryName"], r["aup"]);
            }
        }

        Separator();
    }

    // If a relationship is not available for a given scenario, or an ad-hoc
    // join is needed for some other kind of query, regular INNER/OUTER joins
    // are available in Entity SQL. This example does the same as RelationshipTraversal()
    // but using an INNER JOIN
    static void RegularInnerJoin()
    {
        using (EntityConnection con = new EntityConnection(CONNECTION_STRING))
        {
            con.Open();

            // Find all products from the "Beverages" category
            // NOTE: the ExeuteReader call below is commented out, because this sample
            // won't run against the sample database; this is because the sample EDM
            // schema is modeled using EDM relationships, not primary/foreign keys, so
            // there is no CategoryID property in the Product entity.
            EntityCommand cmd = new EntityCommand(
                "SELECT VALUE p " +
                "FROM Northwind.Products AS p " +
                "INNER JOIN Northwind.Categories AS c " +
                    "ON p.CategoryID = c.CategoryID " +
                "WHERE c.CategoryName = @name",
                con );
            cmd.Parameters.AddWithValue("name", "Beverages");

            // Entity command requires SequentialAccess
            ////DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            ////while(r.Read()) {
            ////   Console.WriteLine("{0}", r["ProductName"]);
            ////}
        }

        Separator();
    }

    // Obtain information about categories and products by formulating a 
    // query that returns a nested result-set containing both Category and
    // Product instances
    static void NestedResultSets() {
        using (EntityConnection con = new EntityConnection(CONNECTION_STRING))
        {
            con.Open();

            // In Entity SQL you can project out a collection (such as the results
            // of a nested subquery) and they show up as nested results in the resulting
            // datareader. Also, note that if we only used the result of NAVIGATE we would
            // only receive the references (key information) to the nested items; the DEREF
            // operator de-references the references into the actual entities.
            EntityCommand cmd = new EntityCommand(
                "SELECT c, (SELECT VALUE DEREF(p) " +
                            "FROM NAVIGATE(c, NorthwindLib.Category_Product) AS p) AS Products " +
                "FROM Northwind.Categories AS c",
                con);
            
            // Entity command requires SequentialAccess
            DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while(r.Read()) {
                // the top-level result-set has two columns, one contains the Category entity and
                // the other contains a datareader containing all the products that we selected for
                // that entity (which, based on the query will be the products associated with that
                // particular category)

                // Get the category record and print out the category name
                DbDataRecord categoryRec = (DbDataRecord)r.GetValue(0);
                Console.WriteLine(categoryRec["CategoryName"]);

                // Now iterate over the products for this category row
                DbDataReader products = r.GetData(1);
                while(products.Read()) {
                    Console.WriteLine("\t" + products["ProductName"]);
                }
            }
        }

        Separator();
    }

    static void Aggregate() {
        using(EntityConnection con = new EntityConnection(CONNECTION_STRING)) {
            con.Open();

            // In Entity SQL you can project out a collection (such as the results
            // of a nested subquery) and they show up as nested results in the resulting
            // datareader. Also, note that if we only used the result of NAVIGATE we would
            // only receive the references (key information) to the nested items; the DEREF
            // operator de-references the references into the actual entities.
            EntityCommand cmd = new EntityCommand(
                "SELECT Name, SqlServer.COUNT(1) AS Count " +
                "FROM Northwind.Products AS p " +
                "GROUP BY NAVIGATE(p, NorthwindLib.Category_Product).CategoryName AS Name",
                con);
            
            // Entity command requires SequentialAccess
            DbDataReader r = cmd.ExecuteReader(CommandBehavior.SequentialAccess);

            while(r.Read()) {
                Console.WriteLine("{0}\t{1}", r[0], r[1]);
            }
        }

        Separator();
    }

    static void Separator() {
        Console.WriteLine("------------------------------------------------");
    }

}
